@@ -21,8 +21,6 @@
 #pragma once
 
 #include <webkitdom/webkitdom.h>
-#define WEBKIT_DOM_USE_UNSTABLE_API
-#include <webkitdom/WebKitDOMDOMSelection.h>
 
 G_BEGIN_DECLS
 
@@ -39,6 +37,8 @@ gboolean ephy_web_dom_utils_find_form_auth_elements (WebKitDOMHTMLFormElement *f
                                                      WebKitDOMNode           **username,
                                                      WebKitDOMNode           **password);
 
+gboolean ephy_web_dom_utils_form_contains_sensitive_element (WebKitDOMHTMLFormElement *form);
+
 void ephy_web_dom_utils_get_absolute_bottom_for_element (WebKitDOMElement *element,
                                                          double           *x,
                                                          double           *y);
